#include "libxml/globals.h"
#include "libxml/valid.h"
#include "libxml/tree.h"

#include <vector>
#include <string>
#include <map>
#include <sstream>
#include <fstream>
#include <iostream>
#include <stdlib.h>
#include <cmath>

#include <globus_common.h>
#include <globus_rsl.h>
extern "C" {
#include <globus_rls_client.h>
}

#include "ngui.h"
#include "ngexec.h"

#define REALSI2000PERMHZ 28.5
#define SI2000PERMHZ 7

using namespace std;

static bool debug = true;

struct executorData {
  string jobDefID;
  string jobName;
  string jobExeID;
};

struct jobExeData {
  string jobDefID;
  string jobName;
  string jobExeID;
  string infoExecutor;
  string action;
};

struct cpuConsumption {
  string count;
  string unit;
};

struct diskConsumption {
  string count;
  string unit;
};

struct minimumRAM {
  string count;
  string unit;
};

struct dataset {
  string name;
};

struct fileInfo {
  string stream;
  string LFN;
  string type;
  string logCol;
  dataset data;
  vector <string> turlList;
  vector <string> SEList;
};

struct actualPar {
  string name;
  int position;
  string type;
  string value;
  string metaType;
};

struct jobDef {
  string jobDefID;
  string transImpl;
  string transUses;
  string transHomePackage;
  string transVersion;
  string jobName;
  cpuConsumption CPU;
  diskConsumption disk;
  minimumRAM ram;
  string ipconnectivity;
  vector <fileInfo> jobLogs;
  vector <fileInfo> jobInputs;
  vector <fileInfo> jobOutputs;
  vector <actualPar> jobPars;
};

struct fileMetaData {
  string lfn;
  string lcn;
  string md5sum;
  string date;
  string size;
  string guid;
  string dataset;
};


static void GetLine (string & s, string & line) {

  int i = s.find ('\n');
  if (i == string::npos) return;

  line = s.substr (0, i);
  s = s.substr (i + 1);
}


static string GetEnv (const string & env) {

  const char * temp = getenv (env.c_str());
  return temp ? temp : "";
}


static int Minutes (const string & period) {

  int minutes = 0;

  int pos = string::npos;
  int len;

  for (int i = 0; i != period.length(); i++) {
    if (isdigit (period[i])) {
      if (pos == string::npos) {pos = i; len = 0;}
      len ++;
    }
    else if (pos != string::npos) {
      switch (period[i]) {
      case 'w':
      case 'W':
        minutes += atoi (period.substr (pos, len).c_str()) * 60 * 24 * 7;
        pos = string::npos;
        break;
      case 'd':
      case 'D':
        minutes += atoi (period.substr (pos, len).c_str()) * 60 * 24;
        pos = string::npos;
        break;
      case 'h':
      case 'H':
        minutes += atoi (period.substr (pos, len).c_str()) * 60;
        pos = string::npos;
        break;
      case 'm':
      case 'M':
        minutes += atoi (period.substr (pos, len).c_str());
        pos = string::npos;
        break;
      case ' ':
        break;
      default:
        return -1;
        break;
      }
    }
  }
  if (pos != string::npos)
    minutes += atoi (period.substr (pos, len).c_str());
  return minutes;
}


string FileToString (const string & filename) {

  ifstream file (filename.c_str());
  if (!file) return "";
  file.seekg (0, ios::end);
  int length = file.tellg();
  file.seekg (0, ios::beg);
  char * buffer = new char [length + 1];
  file.read (buffer, length);
  buffer [length] = 0;
  file.close();
  string content = buffer;
  delete buffer;
  return content;
}


string WeirdWindmillDate (const string & date) {

  return date.substr (11) + '-' +
    date.substr (8, 2) + '-' +
    date.substr (5, 2) + '-' +
    date.substr (0, 4);
}


int FindRelation (const string & attr, globus_rsl_t ** relation,
		  globus_rsl_t * axrsl) {

  if (globus_rsl_is_boolean (axrsl)) {
    globus_list_t * rlist = globus_rsl_boolean_get_operand_list (axrsl);
    while (!globus_list_empty (rlist)) {
      if (FindRelation (attr, relation,
			(globus_rsl_t *) globus_list_first (rlist)))
	return 1;
      rlist = globus_list_rest (rlist);
    }
    return 0;
  }
  else if (globus_rsl_is_relation (axrsl)) {
    if (globus_rsl_is_relation_attribute_equal (axrsl,
						(char *) attr.c_str())) {
      if (*relation) {
	cerr << "Error: Multiple definitions of XRSL attribute \"" << attr
	     << "\"" << endl;
	return 1;
      }
      *relation = axrsl;
    }
    return 0;
  }
  else {
    cerr << "Error: Unexpected XRSL token" << endl;
    return 1;
  }
}


char * numJobsWanted (char * xml) {

  if (debug)
    cout << "xml passed by supervisor: " << xml << endl;

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

  cpuConsumption cpu;
  diskConsumption disk;
  minimumRAM ram;
  string transUses;
  string ipConnectivity;

  while (xmlnode1) {
    if (strcmp ((const char *) xmlnode1->name, "numJobsWanted") == 0) {
      xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
      while (xmlnode2) {
	if (strcmp ((const char *) xmlnode2->name, "minimumResources") == 0) {
	  xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	  while (xmlnode3) {
	    if (strcmp ((const char *) xmlnode3->name, "transUses") == 0) {
	      if (xmlnode3->xmlChildrenNode)
		transUses = (const char *) xmlnode3->xmlChildrenNode->content;
	    }
	    if (strcmp ((const char *) xmlnode3->name, "cpuConsumption") == 0) {
	      xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
	      while (xmlnode4) {
		if (strcmp ((const char *) xmlnode4->name, "count") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    cpu.count = (const char *) xmlnode4->xmlChildrenNode->content;
		if (strcmp ((const char *) xmlnode4->name, "unit") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    cpu.unit = (const char *) xmlnode4->xmlChildrenNode->content;
		xmlnode4 = xmlnode4->next;
	      }
	    }
	    if (strcmp ((const char *) xmlnode3->name, "diskConsumption") == 0) {
	      xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
	      while (xmlnode4) {
		if (strcmp ((const char *) xmlnode4->name, "count") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    disk.count = (const char *) xmlnode4->xmlChildrenNode->content;
		if (strcmp ((const char *) xmlnode4->name, "unit") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    disk.unit = (const char *) xmlnode4->xmlChildrenNode->content;
		xmlnode4 = xmlnode4->next;
	      }
	    }
	    if (strcmp ((const char *) xmlnode3->name, "ipConnectivity") == 0) {
	      if (xmlnode3->xmlChildrenNode)
		ipConnectivity = (const char *) xmlnode3->xmlChildrenNode->content;
	    }
	    if (strcmp ((const char *) xmlnode3->name, "mimimumRAM") == 0) {
	      xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
	      while (xmlnode4) {
		if (strcmp ((const char *) xmlnode4->name, "count") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    ram.count = (const char *) xmlnode4->xmlChildrenNode->content;
		if (strcmp ((const char *) xmlnode4->name, "unit") == 0)
		  if (xmlnode4->xmlChildrenNode)
		    ram.unit = (const char *) xmlnode4->xmlChildrenNode->content;
		xmlnode4 = xmlnode4->next;
	      }
	    }
	    xmlnode3 = xmlnode3->next;
	  }
	}
	xmlnode2 = xmlnode2->next;
      }
    }
    xmlnode1 = xmlnode1->next;
  }

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  int njob = 0;

  stringstream stream, dummy;
  streambuf * sbo = cout.rdbuf(stream.rdbuf());
  streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
  vector <string> e;
  vector <string> r;
  string dulrej = FileToString (GetEnv ("HOME") + "/.dulcineareject");
  while (!dulrej.empty()) {
    string rej;
    GetLine (dulrej, rej);
    r.push_back (rej);
  }    
  ngstatxx (e, e, e, r, e, false, e, true, true, 20, 0, true);
  cerr.rdbuf(sbe);
  cout.rdbuf(sbo);

  string s = stream.str();

  string AtlasUses;
  if (!transUses.empty()) {
    int i = -1;
    do {
      int j = transUses.find (' ', i + 1);
      string sub;
      if (j != string::npos)
	sub = transUses.substr (i + 1, j - i - 1);
      else
	sub = transUses.substr (i + 1);
      if (strncmp (sub.substr (0, 6).c_str(), "Atlas-", 6) == 0) {
	AtlasUses = "atlas-" + sub.substr (6) + ".0";
      }
      i = j;
    } while (i != string::npos);
  }

  if (!AtlasUses.empty()) transUses = AtlasUses;

  if( debug )
    cout << "transUses: " << transUses << endl;

  while (s.size()>0) {
    string line;
    int clusterfreq;
    int clusterncpu;
    int clusterucpu;
    int clusternjob;
    int clusterqjob;
    int clustermem;
    int queuefreq;
    int queuencpu;
    int queueucpu;
    int queuenjob;
    int queueqjob;
    int queuemem;
    int maxtime;
    int mintime;
    int maxuserrun;
    int userfree;
    int userqueue;
    long long userdisk;
    vector <string> rte;

    GetLine (s, line);

    if (line.substr (0, 7) == "Cluster") {
      clusterfreq = 1000;
      clusterncpu = UNDEFINED;
      clusterucpu = UNDEFINED;
      clusternjob = UNDEFINED;
      clusterqjob = UNDEFINED;
      clustermem = UNDEFINED;
      rte.clear();
      while (s.size()>0) {
	GetLine (s, line);
	if (line.substr (0, 2) != "  ") break;
	int cpos = line.find (':', 2);
	if (cpos != string::npos) {
	  string key = line.substr (2, cpos - 2);
	  string val;
	  if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	  if (key == "Installed runtime environments") {
	    while (s.size()>0) {
	      GetLine (s, line);
	      if (line.substr (0, 4) != "    ") break;
	      rte.push_back (line.substr (4));
	    }
	    if (line.substr (0, 2) != "  ") break;
	    cpos = line.find (':', 2);
	    if (cpos != string::npos) {
	      key = line.substr (2, cpos - 2);
	      if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	    }
	  }
	  if (key == "CPU") {
	    int pos = val.find (" MHz");
	    if (pos != string::npos) {
	      int numpos = val.rfind (' ', pos - 1) + 1;
	      clusterfreq = atoi (val.substr (numpos, pos - numpos).c_str());
	    }
	  }
	  if (key == "Number of CPUs")
	    clusterncpu = atoi (val.c_str());
	  if (key == "Number of used CPUs")
	    clusterucpu = atoi (val.c_str());
	  if (key == "Number of jobs")
	    clusternjob = atoi (val.c_str());
	  if (key == "Number of queued jobs")
	    clusterqjob = atoi (val.c_str());
	  if (key == "Memory on each node")
	    clustermem = atoi (val.c_str());
	}
      }
    }

    if (line.substr (0, 5) == "Queue") {
      queuefreq = clusterfreq;
      queuencpu = clusterncpu;
      queueucpu = clusterucpu;
      queuenjob = clusternjob;
      queueqjob = clusterqjob;
      queuemem = clustermem;
      mintime = UNDEFINED;
      maxtime = UNDEFINED;
      maxuserrun = UNDEFINED;
      userfree = UNDEFINED;
      userqueue = UNDEFINED;
      userdisk = UNDEFINED;
      while (s.size()>0) {
	GetLine (s, line);
	if (line.substr (0, 2) != "  ") break;
	int cpos = line.find (':', 2);
	if (cpos != string::npos) {
	  string key = line.substr (2, cpos - 2);
	  string val;
	  if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	  if (key == "User number of free CPUs") {
	    while (s.size()>0) {
	      GetLine (s, line);
	      if (line.substr (0, 4) != "    ") break;
	      if (line.find ("None") != string::npos)
		userfree = 0;
	      else if (line.find ("for infinite time") != string::npos) {
		int ppos = line.find (" processor");
		if (ppos != string::npos) {
		  int free = atoi (line.substr (4, ppos - 4).c_str());
		  if (userfree == UNDEFINED || userfree < free)
		    userfree = free;
		}
	      }
	      else {
		int fpos = line.find ("for ");
		int scale = 1;
		if (cpu.unit == "specint2000days")
		  scale = 86400;
		else if (cpu.unit == "specint2000hours")
		  scale = 3600;
		else if (cpu.unit == "specint2000minutes")
		  scale = 60;
		else if (cpu.unit != "specint2000seconds")
		  cerr << "Unexpected unit " << cpu.unit
		       << ". Assuming specint2000seconds." << endl;
		if (fpos != string::npos && Minutes (line.substr (fpos + 4)) >
		    atoll (cpu.count.c_str()) * scale / SI2000PERMHZ / queuefreq) {
		  int ppos = line.find (" processor");
		  if (ppos != string::npos) {
		    int free = atoi (line.substr (4, ppos - 4).c_str());
		    if (userfree == UNDEFINED || userfree < free)
		      userfree = free;
		  }
		}
	      }
	    }
	    if (line.substr (0, 2) != "  ") break;
	    cpos = line.find (':', 2);
	    if (cpos != string::npos) {
	      key = line.substr (2, cpos - 2);
	      if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	    }
	  }
	  if (key == "CPU") {
	    int pos = val.find (" MHz");
	    if (pos != string::npos) {
	      int numpos = val.rfind (' ', pos - 1) + 1;
	      queuefreq = atoi (val.substr (numpos, pos - numpos).c_str());
	    }
	  }
	  if (key == "Number of CPUs")
	    queuencpu = atoi (val.c_str());
	  if (key == "Number of used CPUs")
	    queueucpu = atoi (val.c_str());
	  if (key == "Number of jobs")
	    queuenjob = atoi (val.c_str());
	  if (key == "Number of queued jobs")
	    queueqjob = atoi (val.c_str());
	  if (key == "Memory on each node")
	    queuemem = atoi (val.c_str());
	  if (key == "Min CPU time")
	    mintime = Minutes (val);
	  if (key == "Max CPU time")
	    maxtime = Minutes (val);
	  if (key == "Max number of running jobs per local user")
	    maxuserrun = atoi (val.c_str());
	  if (key == "User queue length")
	    userqueue = atoi (val.c_str());
	  if (key == "User disk space")
	    userdisk = atoll (val.c_str());
	}
      }

      if (!transUses.empty()) {
	bool found = false;
	for (vector <string>::iterator vsi = rte.begin(); !found && vsi != rte.end(); vsi++) {
	  if (strcasecmp (vsi->c_str(), transUses.c_str()) == 0) found = true;
	}
	if (!found) continue;
      }

      long long rdisk = atoll (disk.count.c_str());
      if      (disk.unit == "KB") rdisk *= 1024;
      else if (disk.unit == "MB") rdisk *= 1024 * 1024;
      else if (disk.unit == "GB") rdisk *= 1024 * 1024 * 1024;

      long long rmem = atoll (ram.count.c_str());
      if      (ram.unit == "KB") rmem *= 1024;
      else if (ram.unit == "MB") rmem *= 1024 * 1024;
      else if (ram.unit == "GB") rmem *= 1024 * 1024 * 1024;

      int scale = 1;
      if (cpu.unit == "specint2000days")
	scale = 86400;
      else if (cpu.unit == "specint2000hours")
	scale = 3600;
      else if (cpu.unit == "specint2000minutes")
	scale = 60;
      else if (cpu.unit != "specint2000seconds")
	cerr << "Unexpected unit " << cpu.unit << ". Assuming specint2000seconds." << endl;

      if (rmem <= queuemem && userfree != UNDEFINED &&
	  (mintime == UNDEFINED ||
	   atoll (cpu.count.c_str()) * scale / SI2000PERMHZ / queuefreq >= mintime) &&
	  (maxtime == UNDEFINED ||
	   atoll (cpu.count.c_str()) * scale / SI2000PERMHZ / queuefreq <= maxtime) &&
	  queueucpu + queueqjob < 2 * queuencpu)
	njob += std::min (userdisk / rdisk, (long long) 2 * queuencpu - queueucpu - queueqjob);
    }
  }

  stringstream conv;
  conv << njob;
  string njobs;
  njobs = conv.str();

  xmlDocPtr doc = xmlNewDoc (BAD_CAST "1.0");

  xmlCreateIntSubset (doc, BAD_CAST "windmill", BAD_CAST "SYSTEM",
		      BAD_CAST "exe_numJobsWanted.dtd");

  xmlNodePtr root = xmlNewDocNode (doc, NULL, BAD_CAST "windmill", NULL);
  xmlDocSetRootElement (doc, root);
  xmlNewProp (root, BAD_CAST "version", BAD_CAST "0.7");
  xmlNewProp (root, BAD_CAST "user", BAD_CAST "executor");
  xmlNewProp (root, BAD_CAST "type", BAD_CAST "respond");

  xmlNodePtr njw = xmlNewChild (root, NULL, BAD_CAST "numJobsWanted", NULL);
  xmlNodePtr avr = xmlNewChild (njw, NULL, BAD_CAST "availableResources", NULL);
  xmlNewChild (avr, NULL, BAD_CAST "jobCount", BAD_CAST njobs.c_str());
  xmlNodePtr cpun = xmlNewChild (avr, NULL, BAD_CAST "cpuMax", NULL);
  xmlNewChild (cpun, NULL, BAD_CAST "count", BAD_CAST cpu.count.c_str());
  xmlNewChild (cpun, NULL, BAD_CAST "unit", BAD_CAST cpu.unit.c_str());

  xmlChar * x;
  int len;

  xmlDocDumpFormatMemoryEnc (doc, &x, &len, "utf-8", 0);

  xmlFreeDoc (doc);

  if (debug)
    cout << "xml passed by executor: " << x << endl;

  return (char *) x;
}

char * executeJobs (char * xml) {

  if (debug)
    cout << "xml passed by supervisor: " << xml << endl;

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode = xmlroot->xmlChildrenNode;

  vector <jobDef> vJobDef;

  while (xmlnode) {
    if (strcmp ((const char*) xmlnode->name, "executeJobs") == 0) {
      xmlNodePtr xmlnode1 = xmlnode->xmlChildrenNode;
      while (xmlnode1) {
	if (strcmp ((const char*) xmlnode1->name, "jobDef") == 0) {
	  xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;

	  jobDef job;

	  while (xmlnode2) {
	    if (strcmp ((const char*) xmlnode2->name, "jobDefID") == 0) {
	      if (xmlnode2->xmlChildrenNode)
		job.jobDefID = (const char*) xmlnode2->xmlChildrenNode->content;
	    }
	    if (strcmp ((const char*) xmlnode2->name, "jobName") == 0) {
	      if (xmlnode2->xmlChildrenNode)
		job.jobName = (const char*) xmlnode2->xmlChildrenNode->content;
	    }
	    if (strcmp ((const char*) xmlnode2->name, "transImplementation") == 0) {
	      if (xmlnode2->xmlChildrenNode)
		job.transImpl = (const char*) xmlnode2->xmlChildrenNode->content;
	    }
	    if (strcmp ((const char*) xmlnode2->name, "transUses") == 0) {
	      if(xmlnode2->xmlChildrenNode)
		job.transUses = (const char*) xmlnode2->xmlChildrenNode->content;
	    }
	    if (strcmp ((const char*) xmlnode2->name, "transHomePackage") == 0) {
	      if(xmlnode2->xmlChildrenNode)
		job.transHomePackage = (const char*) xmlnode2->xmlChildrenNode->content;
	    }
	    if (strcmp ((const char*) xmlnode2->name, "expectedResources") == 0) {
	      xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	      while (xmlnode3) {
		if (strcmp ((const char*) xmlnode3->name, "cpuConsumption") == 0) {
		  xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
		  while (xmlnode4) {
		    if (strcmp ((const char*) xmlnode4->name, "count") == 0)
		      if (xmlnode4->xmlChildrenNode)
			job.CPU.count = (const char*) xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*) xmlnode4->name, "unit") == 0)
		      if (xmlnode4->xmlChildrenNode)
			job.CPU.unit = (const char*) xmlnode4->xmlChildrenNode->content;
		    xmlnode4 = xmlnode4->next;
		  }
		}
		if (strcmp ((const char*) xmlnode3->name, "diskConsumption") == 0) {
		  xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
		  while (xmlnode4) {
		    if (strcmp ((const char*) xmlnode4->name, "count") == 0)
		      if (xmlnode4->xmlChildrenNode)
			job.disk.count = (const char*) xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*) xmlnode4->name, "unit") == 0)
		      if(xmlnode4->xmlChildrenNode)
			job.disk.unit = (const char*) xmlnode4->xmlChildrenNode->content;
		    xmlnode4 = xmlnode4->next;
		  }
		}
		if (strcmp ((const char*) xmlnode3->name, "minimumRAM") == 0) {
		  xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;
		  while (xmlnode4) {
		    if (strcmp ((const char*) xmlnode4->name, "count") == 0)
		      if(xmlnode4->xmlChildrenNode)
			job.ram.count = (const char*) xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*) xmlnode4->name, "unit") == 0)
		      if(xmlnode4->xmlChildrenNode)
			job.ram.unit = (const char*) xmlnode4->xmlChildrenNode->content;
		    xmlnode4 = xmlnode4->next;
		  }
		}
		if (strcmp ((const char*) xmlnode3->name, "ipConnectivity") == 0) {
		  if(xmlnode3->xmlChildrenNode)
		    job.ipconnectivity = (const char*) xmlnode3->xmlChildrenNode->content;
		}
		xmlnode3 = xmlnode3->next;
	      }
	    }
	    if (strcmp ((const char*) xmlnode2->name, "jobPars") == 0) {
	      xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	      while (xmlnode3) {
		if (strcmp((const char*) xmlnode3->name, "actualPar") == 0) {
		  xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;

		  actualPar pars;

		  while (xmlnode4) {
		    if (strcmp ((const char*) xmlnode4->name, "name") == 0)
		      if (xmlnode4->xmlChildrenNode)
			pars.name = (const char*) xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*) xmlnode4->name, "value") == 0)
		      if (xmlnode4->xmlChildrenNode)
			pars.value = (const char*) xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*) xmlnode4->name, "position") == 0)
		      if (xmlnode4->xmlChildrenNode)
			pars.position = atoi ((const char*) xmlnode4->xmlChildrenNode->content);
		    if (strcmp ((const char*) xmlnode4->name, "metaType") == 0)
		      if (xmlnode4->xmlChildrenNode)
			pars.metaType = (const char*) xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*) xmlnode4->name, "type") == 0)
		      if (xmlnode4->xmlChildrenNode)
			pars.type = (const char*) xmlnode4->xmlChildrenNode->content;
		    xmlnode4 = xmlnode4->next;
		  }
		  job.jobPars.push_back (pars);
		}
		xmlnode3 = xmlnode3->next;
	      }
	    }
	    if ((strcmp ((const char*) xmlnode2->name, "jobInputs") == 0) ||
		(strcmp ((const char*) xmlnode2->name, "jobOutputs") == 0) ||
		(strcmp ((const char*) xmlnode2->name, "jobLogs") == 0)) {
	      xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	      while(xmlnode3) {
		if (strcmp ((const char*) xmlnode3->name, "fileInfo") == 0) {
		  xmlNodePtr xmlnode4 = xmlnode3->xmlChildrenNode;

		  fileInfo info;

		  while(xmlnode4) {
		    if (strcmp ((const char*) xmlnode4->name, "stream") == 0)
		      if (xmlnode4->xmlChildrenNode)
			info.stream = (const char*)xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*) xmlnode4->name, "LFN") == 0)
		      if (xmlnode4->xmlChildrenNode) {
			info.LFN = (const char*) xmlnode4->xmlChildrenNode->content;
			//			if( info.LFN[ info.LFN.size()-1 ] != '1' ) info.LFN += ".1";
		      }
		    if (strcmp ((const char*) xmlnode4->name, "dataset") == 0) {
		      xmlNodePtr xmlnode5 = xmlnode4->xmlChildrenNode;
		      while(xmlnode5) {
			if (strcmp ((const char*) xmlnode5->name, "name") == 0)
			  if (xmlnode5->xmlChildrenNode)
			    info.data.name = (const char*) xmlnode5->xmlChildrenNode->content;
			xmlnode5 = xmlnode5->next;
                      }
                    }
		    if (strcmp ((const char*) xmlnode4->name, "type") == 0)
		      if(xmlnode4->xmlChildrenNode)
			info.type = (const char*) xmlnode4->xmlChildrenNode->content;
		    if (strcmp ((const char*) xmlnode4->name, "logCol") == 0) {
		      if (xmlnode4->xmlChildrenNode)
			info.logCol = (const char*) xmlnode4->xmlChildrenNode->content;
		      if (info.logCol [0] != '/' ) info.logCol = "/" + info.logCol;
		      if (info.logCol [info.logCol.length() - 1] != '/' ) info.logCol += "/";
		    }
		    if (strcmp ((const char*) xmlnode4->name, "SEList") == 0) {
		      xmlNodePtr xmlnode5 = xmlnode4->xmlChildrenNode;
		      while (xmlnode5) {
			if (strcmp ((const char*) xmlnode5->name, "SE") == 0)
			  info.SEList.push_back
			    ((const char*) xmlnode5->xmlChildrenNode->content);
			xmlnode5 = xmlnode5->next;
		      }
		    }
		    if (strcmp ((const char*) xmlnode4->name, "turlList") == 0) {
		      xmlNodePtr xmlnode5 = xmlnode4->xmlChildrenNode;
		      while (xmlnode5) {
			if (strcmp ((const char*) xmlnode5->name, "turl") == 0)
			  info.turlList.push_back
			    ((const char*) xmlnode5->xmlChildrenNode->content);
			xmlnode5 = xmlnode5->next;
		      }
		    }
		    xmlnode4 = xmlnode4->next;
		  }
		  if (!info.LFN.empty()) {
		    if (strcmp ((const char*) xmlnode2->name, "jobInputs") == 0)
		      job.jobInputs.push_back (info);
		    if (strcmp ((const char*) xmlnode2->name, "jobOutputs") == 0)
		      job.jobOutputs.push_back (info);
		    if (strcmp ((const char*) xmlnode2->name, "jobLogs") == 0)
		      job.jobLogs.push_back (info);
		  }
		}
		xmlnode3 = xmlnode3->next;
	      }
	    }
	    xmlnode2 = xmlnode2->next;
	  }
	  vJobDef.push_back (job);
	}
	xmlnode1 = xmlnode1->next;
      }
    }
    xmlnode = xmlnode->next;
  }

  // paths
  string dc2ftp ("ftp://ftp.nordugrid.org/applications/hep/atlas/dc2/");
  string castorgridpath ("castorgridtest.cern.ch/castor/cern.ch/grid/atlas");
  string nordugridrls ("gridsrv3.nbi.dk");

  // constructing xrslstring
  vector<string> xrslstrings;

  globus_module_activate (GLOBUS_COMMON_MODULE);
  globus_module_activate (GLOBUS_IO_MODULE);
  globus_module_activate (GLOBUS_RLS_CLIENT_MODULE);

  for (vector <jobDef>::iterator job = vJobDef.begin();
       job != vJobDef.end(); job++) {

    string xrslstring = "&";

    string use_jobTrans; // Does job use jobTransforms?
    string atlas_release;

    if (!job->transUses.empty()) {
      int i = -1;
      do {
        int j = job->transUses.find (' ', i + 1);
        string sub;
	if (j != string::npos)
	  sub = job->transUses.substr (i + 1, j - i - 1);
	else
	  sub = job->transUses.substr (i + 1);
	if (strncmp (sub.substr (0, 14).c_str(), "JobTransforms-", 14) == 0) {
	  int k = sub.find_first_of ("0123456789", 14);
	  while (k != string::npos) {
	    if (!use_jobTrans.empty()) use_jobTrans += '-';
	    use_jobTrans += "0";
	    use_jobTrans += sub [k];
	    k = sub.find_first_of ("0123456789", k + 1);
	  }
	}
	else {
	  if (strncmp (sub.substr (0, 6).c_str(), "Atlas-", 6) == 0)
	    atlas_release = sub.substr (6);
	  xrslstring += "(runTimeEnvironment=\"" + sub + "\")";
	}
        i = j;
      } while (i != string::npos);
    }

    if (!use_jobTrans.empty() && !atlas_release.empty())
      xrslstring += "(environment=(T_RELEASE " + atlas_release + "))";

    if (!job->transImpl.empty()) {
      if (!use_jobTrans.empty()) {
        xrslstring += "(executable=NGExecWrapper)";
      }
      else {
	if (job->transImpl == "cp")
	  job->transImpl = "/bin/cp";
	else if (job->transImpl == "dd")
	  job->transImpl = "/bin/dd";
        xrslstring += "(executable=\"" + job->transImpl + "\")";
      }
    }

    int maxarg = 0;
    //    vector<string> inputLFNs;

    for (vector <actualPar>::iterator pars = job->jobPars.begin();
	 pars != job->jobPars.end(); pars++) {
      if (maxarg < pars->position) maxarg = pars->position;
    }

    int offset = 0;
    vector<string> guids;

    if (!use_jobTrans.empty()) {
      offset = 3 + 2 * job->jobInputs.size() + 3 * (job->jobOutputs.size() + job->jobLogs.size());

      // grap GUID's for inputFiles
      string rlsserver("rls://");
      rlsserver += nordugridrls;

      globus_result_t res;
      globus_rls_handle_t* handle;
      int errc;
      char errmsg[512];

      res = globus_rls_client_connect( (char*)rlsserver.c_str(), &handle );
      if (res) {
	globus_rls_client_error_info (res, &errc, errmsg, 512, false);
	cout << errmsg << endl;
	continue; // cannot submit job then
      }

      for( vector<fileInfo>::iterator inputfile = job->jobInputs.begin(); inputfile != job->jobInputs.end(); inputfile++ ) {
	
	globus_list_t* attr_list;
	res = globus_rls_client_lrc_attr_value_get( handle, (char*)(inputfile->LFN.c_str()), NULL, globus_rls_obj_lrc_lfn, &attr_list);

	if( res ) {
	  globus_rls_client_error_info (res, &errc, errmsg, 512, false);
	  cout << errmsg << endl;
	  continue;
	}
	
	for( globus_list_t* list = attr_list; list; list = globus_list_rest(list) ) {

	  globus_rls_attribute_t* attr = (globus_rls_attribute_t*)globus_list_first( list );
	  if( strcmp( attr->name, "dq_guid" ) != 0) continue;

	  guids.push_back( attr->val.s );
	  break; // found dq_guid
	}

      }

      res = globus_rls_client_close (handle);
      if (res) {
	globus_rls_client_error_info (res, &errc, errmsg, 512, false);
	cout << errmsg << endl;
      }

      if( guids.size() != job->jobInputs.size() ) {
	cout << "Not all guids for jobInputs were found" << endl;
	continue; // cannot submit job then
      }
    }

    vector <string> arguments (maxarg + offset);

    for (vector <actualPar>::iterator pars = job->jobPars.begin();
	 pars != job->jobPars.end(); pars++) {
      if (strncmp (pars->metaType.c_str(), "namedPar", 8) == 0 )
	arguments [pars->position - 1 + offset] = pars->name + "=" + pars->value;
      else
	arguments [pars->position - 1 + offset] = pars->value;
    }

    if (!use_jobTrans.empty()) {
      arguments[0] = job->transImpl;

      stringstream ss1;
      ss1 << job->jobInputs.size();
      arguments[1] = ss1.str();
      for (int i = 0; i < job->jobInputs.size(); i++ ) {
	arguments[2 + 2*i] = job->jobInputs [i].LFN;
	arguments[2 + 2*i + 1] = guids[i];
      }

      stringstream ss2;
      ss2 << ( job->jobOutputs.size() + job->jobLogs.size() );
      arguments[2 + 2*job->jobInputs.size() ] = ss2.str();
      for (int i = 0; i < job->jobOutputs.size(); i++) {
	arguments [3 + 2*job->jobInputs.size() + 3*i] = job->jobOutputs [i].LFN;
	arguments [3 + 2*job->jobInputs.size() + 3*i + 1] = job->jobOutputs [i].logCol;
	arguments [3 + 2*job->jobInputs.size() + 3*i + 2] = job->jobOutputs [i].data.name;
      }
      for (int i = 0; i < job->jobLogs.size(); i++) {
	arguments [3 + 2*job->jobInputs.size() + 3*job->jobOutputs.size() + 3*i] = job->jobLogs [i].LFN;
	arguments [3 + 2*job->jobInputs.size() + 3*job->jobOutputs.size() + 3*i + 1] = job->jobLogs [i].logCol;
	arguments [3 + 2*job->jobInputs.size() + 3*job->jobOutputs.size() + 3*i + 2] = job->jobLogs [i].data.name;
      }
    }

    if (!arguments.empty()) {
      xrslstring += "(arguments=";
      for (unsigned int j = 0; j < maxarg + offset; j++ ) {
	xrslstring += "\"" + arguments[j] + "\" ";
      }
      xrslstring.erase (xrslstring.length() - 1);
      xrslstring += ")";
    }

    string stdout;
    string stderr;
    string join;
    if (job->jobLogs.size() == 1)
      join = "yes";
    for (int i = 0; i < job->jobLogs.size(); i++) {
      if (strcmp (job->jobLogs[i].stream.c_str(), "stdout") == 0)
        stdout = job->jobLogs[i].LFN;
      if (strcmp (job->jobLogs[i].stream.c_str(), "stderr" ) == 0 )
        stderr = job->jobLogs[i].LFN;
      if (strcmp (job->jobLogs[i].stream.c_str(), "stdboth") == 0 ) {
        stdout = job->jobLogs[i].LFN;
        join = "yes";
      }
    }

    if (!stdout.empty())
      xrslstring += "(stdout=\"" + stdout + "\")";
    if (!stderr.empty())
      xrslstring += "(stderr=\"" + stderr + "\")";
    if (!join.empty())
      xrslstring += "(join=yes)";

    xrslstring += "(gmlog=\"gmlog\")";     	

    if (!job->disk.count.empty() && !job->disk.unit.empty()) {
      string size;
      stringstream stream;
      if (strncmp (job->disk.unit.c_str(), "KB", 2) == 0) {
	long newsize = 1 + atoll (job->disk.count.c_str()) / 1000;
	stream << newsize;
	size = stream.str();
      }
      else if (strncmp (job->disk.unit.c_str(), "GB", 2) == 0) {
	long newsize = atoll (job->disk.count.c_str()) * 1000;
	stream << newsize;
	size = stream.str();
      }
      else {
	if (strncmp (job->disk.unit.c_str(), "MB", 2) != 0)
	  cerr << "Unsupported disk-unit: " << job->disk.unit << ". Assuming MB." << endl;
	size = job->disk.count;
      }
      xrslstring += "(disk=" + size + ")";
    }

    if (!job->ram.count.empty() && !job->ram.unit.empty()) {
      string size;
      stringstream stream;
      if (strncmp (job->ram.unit.c_str(), "KB", 2) == 0) {
	long newsize = 1 + atoll (job->ram.count.c_str()) / 1000;
	stream << newsize;
	size = stream.str();
      }
      else if (strncmp (job->ram.unit.c_str(), "GB", 2) == 0) {
	long newsize = atoll (job->ram.count.c_str()) * 1000;
	stream << newsize;
	size = stream.str();
      }
      else {
	if (strncmp (job->ram.unit.c_str(), "MB", 2) != 0)
	  cerr << "Unsupported ram-unit: " << job->ram.unit << ". Assuming MB." << endl;
	size = job->ram.count;
      }
      xrslstring += "(memory=" + size + ")";
    }

    if (strncmp (job->ipconnectivity.c_str(), "yes", 3) == 0)
      xrslstring += "(nodeAccess=outbound)";

    if (!job->CPU.count.empty() && !job->CPU.unit.empty()) {
      xrslstring += "(GridTime=\"";
      stringstream timestring;
      int scale = 1;
      if (job->CPU.unit == "specint2000days")
	scale = 86400;
      else if (job->CPU.unit == "specint2000hours")
	scale = 3600;
      else if (job->CPU.unit == "specint2000minutes")
	scale = 60;
      else if (job->CPU.unit != "specint2000seconds")
	cerr << "Unexpected unit " << job->CPU.unit << ". Assuming specint2000seconds." << endl;
      timestring << (int) (atoll (job->CPU.count.c_str()) * scale / (SI2000PERMHZ*2800) + 1);
      xrslstring += timestring.str() + "\")";
    }

    if (!job->jobName.empty())
      xrslstring += "(jobName=\"" + job->jobName + "\")";

    if (!job->jobInputs.empty() || !use_jobTrans.empty()) {
      xrslstring += "(inputFiles=";
      for (unsigned int j = 0; j < job->jobInputs.size(); j++) {
	if (!job->jobInputs [j].turlList.empty()) {
	  xrslstring += "(\"" + job->jobInputs[j].LFN + "\" \"" +
	    job->jobInputs[j].turlList[0]+ "\")";
	} else {
	  xrslstring += "(\"" + job->jobInputs[j].LFN + "\"" + " \"rls://" + nordugridrls + "/" + job->jobInputs[j].LFN + "\")";
	}
      }
      if (!use_jobTrans.empty()) {
	xrslstring += "(\"JobTransforms.tar.gz\" \"" + dc2ftp + "JobTransforms-" +
	  use_jobTrans + ".tar.gz\")";
	xrslstring += "(\"NGExecWrapper\" \"" + dc2ftp + "NGExecWrapper\")";
      }
      xrslstring += ")";
    }

    if (!job->jobOutputs.empty() || !job->jobLogs.empty() || !use_jobTrans.empty())
      xrslstring += "(outputFiles=";

    if (!job->jobOutputs.empty()) {
      for (unsigned int j = 0; j < job->jobOutputs.size(); j++) {
	xrslstring += "(\"" + job->jobOutputs[j].LFN + "\" \"rls://";
	if (!job->jobOutputs[j].SEList.empty()) {
	  for (unsigned int k = 0; k < job->jobOutputs[j].SEList.size(); k++) {
	    xrslstring += "gsiftp://";
	    if (strncmp (job->jobOutputs[j].SEList[k].c_str(), "wacdr002d.cern.ch", 17) == 0 ||
		strncmp (job->jobOutputs[j].SEList[k].c_str(), "castorgrid.cern.ch", 18) == 0 ) {
	      string logcol (job->jobOutputs[j].logCol);
	      if (logcol [logcol.length() - 1] == '/') logcol.resize (logcol.length() - 1);
	      xrslstring += castorgridpath + logcol;
	    }
	    else
	      xrslstring += job->jobOutputs[j].SEList[k];
	    if (k != job->jobOutputs[j].SEList.size() - 1) xrslstring += "|";
	    }
	  }
	xrslstring += "@" + nordugridrls + '/' + job->jobOutputs[j].LFN + "\")";
      }
    }

    if (!job->jobLogs.empty()) {
      for (unsigned int j = 0; j < job->jobLogs.size(); j++) {
	xrslstring += "(\"" + job->jobLogs[j].LFN + "\" \"rls://";
        if (!job->jobLogs[j].SEList.empty()) {
          for (unsigned int k = 0; k < job->jobLogs[j].SEList.size(); k++) {
            xrslstring += "gsiftp://";
	    if (strncmp (job->jobLogs[j].SEList[k].c_str(), "wacdr002d.cern.ch", 17) == 0 ||
		strncmp (job->jobLogs[j].SEList[k].c_str(), "castorgrid.cern.ch", 18) == 0 ) {
	      string logcol (job->jobLogs[j].logCol);
	      if (logcol [logcol.length() - 1] == '/') logcol.resize (logcol.length() - 1);
	      xrslstring += castorgridpath + logcol;
	    }
	    else
	      xrslstring += job->jobLogs[j].SEList[k];
	    if (k != job->jobLogs[j].SEList.size() - 1) xrslstring += "|";
          }
	}
        xrslstring += "@" + nordugridrls + '/' + job->jobLogs[j].LFN + "\")";
      }
    }

    if (!use_jobTrans.empty()) {
      xrslstring += "(\"OutputFiles.xml\" \"\")";
      xrslstring += "(\"metadata.xml\" \"\")";
    }

    if (!job->jobOutputs.empty() || !job->jobLogs.empty() || !use_jobTrans.empty())
      xrslstring += ")";

    if (debug)
      cout << "xrslstring: " << xrslstring << endl;

    xrslstrings.push_back( xrslstring );
  }

  globus_module_deactivate (GLOBUS_RLS_CLIENT_MODULE);
  globus_module_deactivate (GLOBUS_IO_MODULE);
  globus_module_deactivate (GLOBUS_COMMON_MODULE);

  stringstream s, dummy;
  streambuf * sbo = cout.rdbuf(s.rdbuf());
  streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
  vector <string> e;
  string f;
  vector <string> r;
  string dulrej = FileToString (GetEnv ("HOME") + "/.dulcineareject");
  while (!dulrej.empty()) {
    string rej;
    GetLine (dulrej, rej);
    r.push_back (rej);
  }    
  ngsubxx (e, xrslstrings, e, r, e, f, false, false, 20, 2, true);
  cerr.rdbuf(sbe);
  cout.rdbuf(sbo);

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  return NULL;
}

char * getExecutorData (char * xml) {

  if (debug)
    cout << "xml passed by supervisor: " << xml << endl;

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

  vector <executorData> exdata;

  while (xmlnode1) {
    if (strcmp ((const char *) xmlnode1->name, "getExecutorData") == 0) {
      xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
      while (xmlnode2) {
	if (strcmp ((const char *) xmlnode2->name, "jobDef") == 0) {
	  xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	  executorData jobDef;
	  while (xmlnode3) {
	    if (strcmp ((const char *) xmlnode3->name, "jobDefID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobDefID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobName") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobName =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    xmlnode3 = xmlnode3->next;
	  }
	  exdata.push_back (jobDef);
	}
	xmlnode2 = xmlnode2->next;
      }
    }
    xmlnode1 = xmlnode1->next;
  }

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  string filename = GetEnv ("HOME");
  filename.append ("/.ngjobs");
  ifstream ngjobs (filename.c_str());

  for (vector <executorData>::iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++) {
    ngjobs.clear();
    ngjobs.seekg (ios::beg);
    string jobidname;
    while (getline (ngjobs, jobidname)) {
      int pos = jobidname.find('#');
      string jobid = jobidname.substr (0, pos);
      string jobname = jobidname.substr (pos + 1);
      if (jobname == vedi->jobName)
	vedi->jobExeID = jobid;
    }
  }

  ngjobs.close();

  xmlDocPtr doc = xmlNewDoc (BAD_CAST "1.0");

  xmlCreateIntSubset (doc, BAD_CAST "windmill", BAD_CAST "SYSTEM",
		      BAD_CAST "exe_getExecutorData.dtd");

  xmlNodePtr root = xmlNewDocNode (doc, NULL, BAD_CAST "windmill", NULL);
  xmlDocSetRootElement (doc, root);
  xmlNewProp (root, BAD_CAST "version", BAD_CAST "0.7");
  xmlNewProp (root, BAD_CAST "user", BAD_CAST "executor");
  xmlNewProp (root, BAD_CAST "type", BAD_CAST "respond");
  xmlNodePtr ged = xmlNewChild (root, NULL, BAD_CAST "getExecutorData", NULL);

  for (vector <executorData>::const_iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++) {
    xmlNodePtr jed = xmlNewChild (ged, NULL, BAD_CAST "jobExeData", NULL);
    xmlNewChild (jed, NULL, BAD_CAST "jobDefID",
		 BAD_CAST vedi->jobDefID.c_str());
    xmlNewChild (jed, NULL, BAD_CAST "jobName",
		 BAD_CAST vedi->jobName.c_str());
    if (!vedi->jobExeID.empty())
      xmlNewChild (jed, NULL, BAD_CAST "jobExeID",
		   BAD_CAST vedi->jobExeID.c_str());
    xmlNewChild (jed, NULL, BAD_CAST "submitStatus",
		 BAD_CAST (vedi->jobExeID.empty() ? "rejected" : "submitted"));
  }

  xmlChar * x;
  int len;

  xmlDocDumpFormatMemoryEnc (doc, &x, &len, "utf-8", 0);

  xmlFreeDoc (doc);

  if (debug)
    cout << "xml passed by executor: " << x << endl;

  return (char *) x;
}

char * getStatus (char * xml) {

  if (debug)
    cout << "xml passed by supervisor: " << xml << endl;

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

  vector <jobExeData> exdata;

  while (xmlnode1) {
    if (strcmp ((const char *) xmlnode1->name, "getStatus") == 0) {
      xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
      while (xmlnode2) {
	if (strcmp ((const char *) xmlnode2->name, "jobExeData") == 0) {
	  xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	  jobExeData jobDef;
	  while (xmlnode3) {
	    if (strcmp ((const char *) xmlnode3->name, "jobDefID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobDefID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobName") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobName =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobExeID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobExeID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "infoExecutor") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.infoExecutor =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    xmlnode3 = xmlnode3->next;
	  }
	  exdata.push_back (jobDef);
	}
	xmlnode2 = xmlnode2->next;
      }
    }
    xmlnode1 = xmlnode1->next;
  }

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  vector <string> jobs;

  for (vector <jobExeData>::iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++)
    jobs.push_back (vedi->jobExeID);

  stringstream stream, dummy;
  streambuf * sbo = cout.rdbuf(stream.rdbuf());
  streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
  vector <string> e;
  ngstatxx (jobs, e, e, e, e, false, e, false, true, 20, 0, true);
  cerr.rdbuf(sbe);
  cout.rdbuf(sbo);

  map <string, string> cluster;
  map <string, string> queue;
  map <string, string> status;
  map <string, string> error;
  map <string, string> submitted;
  map <string, string> gmlog;
  map <string, string> host;
  map <string, int> cputime;
  map <string, int> walltime;

  string s = stream.str();

  string line;
  while (s.size()>0) {
    if (line.empty() || line.substr (0, 3) != "Job") GetLine (s, line);
    if (line.substr (0, 3) == "Job") {
      string jobid = line.substr (4);
      while (s.size()>0) {
	GetLine (s, line);
	if (line.substr (0, 2) != "  ") break;
	int cpos = line.find (':', 2);
	if (cpos != string::npos) {
	  string key = line.substr (2, cpos - 2);
	  string val;
	  if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	  if (key == "Execution nodes") {
	    while (s.size()>0) {
	      GetLine (s, line);
	      if (line.substr (0, 4) != "    ") break;
	      host [jobid] = line.substr (4);
	      // Need to be improved for parallel jobs, but for Atlas do we need it?
	    }
	    if (line.substr (0, 2) != "  ") break;
	    cpos = line.find (':', 2);
	    if (cpos != string::npos) {
	      key = line.substr (2, cpos - 2);
	      if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	    }
	  }
	  if (key == "Cluster")
	    cluster [jobid] = val;
	  if (key == "Queue")
	    queue [jobid] = val;
	  if (key == "Status")
	    status [jobid] = val;
	  if (key == "Error")
	    error [jobid] = val;
	  if (key == "Submitted")
	    submitted [jobid] = val;
	  if (key == "Used CPU time")
	    cputime [jobid] = Minutes (val);
	  if (key == "Used wall time")
	    walltime [jobid] = Minutes (val);
	  if (key == "Grid manager log directory")
	    gmlog [jobid] = val;
	}
      }
    }
  }

  vector <string> finishedjobs;
  for (vector <jobExeData>::const_iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++)
    if (status [vedi->jobExeID].find ("FINISHED") != string::npos &&
	error [vedi->jobExeID].empty())
      finishedjobs.push_back (vedi->jobExeID);

  if (!finishedjobs.empty()) {
    string dir = "/tmp";
    nggetxx (finishedjobs, e, e, e, e, false, dir, true, true, 20, 0, true);
  }

  map <string, string> xrsl;
  map <string, string> poolfile;
  map <string, map <string, fileMetaData> > metadata;

  for (vector <jobExeData>::const_iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++)
    if (status [vedi->jobExeID].find ("FINISHED") != string::npos) {

      string xrslstr = FileToString ("/tmp/" + vedi->jobName + '/' +
				     gmlog [vedi->jobExeID] + "/description");
      if (!xrslstr.empty()) xrsl [vedi->jobExeID] = xrslstr;

      string poolstr = FileToString ("/tmp/" + vedi->jobName +
				     "/metadata.xml");
      if (!poolstr.empty()) poolfile [vedi->jobExeID] = poolstr;

      string metadataxml = FileToString ("/tmp/" + vedi->jobName +
					 "/OutputFiles.xml");

      system (("rm -rf /tmp/" + vedi->jobName).c_str());

      if (debug)
	cout << "File meta data: " << metadataxml << endl;

      if (!metadataxml.empty ()) {

	xmlInitParser();

	xmlDocPtr xmldoc = xmlParseMemory (metadataxml.c_str(),
					   metadataxml.length());
	if (!xmldoc) return NULL;

	xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

	xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

	map <string, fileMetaData> mfmd;

	while (xmlnode1) {
	  if (strcmp ((const char *) xmlnode1->name, "file") == 0) {
	    xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
	    fileMetaData fmd;
	    while (xmlnode2) {
	      if (strcmp ((const char *) xmlnode2->name, "lfn") == 0)
		if (xmlnode2->xmlChildrenNode)
		  fmd.lfn = (const char *) xmlnode2->xmlChildrenNode->content;
	      if (strcmp ((const char *) xmlnode2->name, "lcn") == 0)
		if (xmlnode2->xmlChildrenNode)
		  fmd.lcn = (const char *) xmlnode2->xmlChildrenNode->content;
	      if (strcmp ((const char *) xmlnode2->name, "md5sum") == 0)
		if (xmlnode2->xmlChildrenNode)
		  fmd.md5sum = (const char *) xmlnode2->xmlChildrenNode->content;
	      if (strcmp ((const char *) xmlnode2->name, "date") == 0)
		if (xmlnode2->xmlChildrenNode)
		  fmd.date = (const char *) xmlnode2->xmlChildrenNode->content;
	      if (strcmp ((const char *) xmlnode2->name, "size") == 0)
		if (xmlnode2->xmlChildrenNode)
		  fmd.size = (const char *) xmlnode2->xmlChildrenNode->content;
	      if (strcmp ((const char *) xmlnode2->name, "guid") == 0)
		if (xmlnode2->xmlChildrenNode)
		  fmd.guid = (const char *) xmlnode2->xmlChildrenNode->content;
	      if (strcmp ((const char *) xmlnode2->name, "dataset") == 0)
		if (xmlnode2->xmlChildrenNode)
		  fmd.dataset = (const char *) xmlnode2->xmlChildrenNode->content;
	      xmlnode2 = xmlnode2->next;
	    }
	    mfmd [fmd.lfn] = fmd;
	  }
	  xmlnode1 = xmlnode1->next;
	}

	xmlFreeDoc (xmldoc);
	xmlCleanupParser();

	metadata [vedi->jobExeID] = mfmd;
      }
    }

  xmlDocPtr doc = xmlNewDoc (BAD_CAST "1.0");

  xmlCreateIntSubset (doc, BAD_CAST "windmill", BAD_CAST "SYSTEM",
		      BAD_CAST "exe_getStatus.dtd");

  xmlNodePtr root = xmlNewDocNode (doc, NULL, BAD_CAST "windmill", NULL);
  xmlDocSetRootElement (doc, root);
  xmlNewProp (root, BAD_CAST "version", BAD_CAST "0.7");
  xmlNewProp (root, BAD_CAST "user", BAD_CAST "executor");
  xmlNewProp (root, BAD_CAST "type", BAD_CAST "respond");
  xmlNodePtr ged = xmlNewChild (root, NULL, BAD_CAST "getStatus", NULL);

  globus_module_activate (GLOBUS_COMMON_MODULE);
  globus_module_activate (GLOBUS_IO_MODULE);
  globus_module_activate (GLOBUS_RLS_CLIENT_MODULE);
  globus_module_activate (GLOBUS_RSL_MODULE);

  for (vector <jobExeData>::const_iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++) {
    xmlNodePtr jed = xmlNewChild (ged, NULL, BAD_CAST "jobExeData", NULL);
    xmlNewChild (jed, NULL, BAD_CAST "jobDefID",
		 BAD_CAST vedi->jobDefID.c_str());
    xmlNewChild (jed, NULL, BAD_CAST "jobName",
		 BAD_CAST vedi->jobName.c_str());
    xmlNewChild (jed, NULL, BAD_CAST "jobExeID",
		 BAD_CAST vedi->jobExeID.c_str());
    if (!vedi->infoExecutor.empty())
      xmlNewChild (jed, NULL, BAD_CAST "infoExecutor",
		   BAD_CAST vedi->infoExecutor.c_str());

    if (!submitted [vedi->jobExeID].empty())
      xmlNewChild (jed, NULL, BAD_CAST "startTime",
		   BAD_CAST WeirdWindmillDate (submitted [vedi->jobExeID]).c_str());
    if (status [vedi->jobExeID].substr (0, 8) == "FINISHED")
      xmlNewChild (jed, NULL, BAD_CAST "endTime",
		   BAD_CAST WeirdWindmillDate (status [vedi->jobExeID].substr (9)).c_str());

    if (cputime [vedi->jobExeID] != 0) {

      stringstream stream, dummy;
      streambuf * sbo = cout.rdbuf(stream.rdbuf());
      streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
      vector <string> e;
      vector <string> c;
      c.push_back (cluster [vedi->jobExeID]);
      ngstatxx (e, e, c, e, e, false, e, true, true, 20, 0, true);
      cerr.rdbuf(sbe);
      cout.rdbuf(sbo);

      string s = stream.str();

      int freq = 1000;

      while (s.size()>0) {
	string line;
	GetLine (s, line);
	if (line.substr (0, 7) == "Cluster" &&
	    line.substr (8) == cluster [vedi->jobExeID]) {
	  while (s.size()>0) {
	    GetLine (s, line);
	    if (line.substr (0, 2) != "  ") break;
	    int cpos = line.find (':', 2);
	    if (cpos != string::npos) {
	      string key = line.substr (2, cpos - 2);
	      string val;
	      if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	      if (key == "CPU") {
		int pos = val.find (" MHz");
		if (pos != string::npos) {
		  int numpos = val.rfind (' ', pos - 1) + 1;
		  freq = atoi (val.substr (numpos, pos - numpos).c_str());
		}
	      }
	    }
	  }
	}

	if (line.substr (0, 5) == "Queue" &&
	    line.substr (6) == queue [vedi->jobExeID]) {
	  while (s.size()>0) {
	    GetLine (s, line);
	    if (line.substr (0, 2) != "  ") break;
	    int cpos = line.find (':', 2);
	    if (cpos != string::npos) {
	      string key = line.substr (2, cpos - 2);
	      string val;
	      if (line.length() > cpos + 2) val = line.substr (cpos + 2);
	      if (key == "CPU") {
		int pos = val.find (" MHz");
		if (pos != string::npos) {
		  int numpos = val.rfind (' ', pos - 1) + 1;
		  freq = atoi (val.substr (numpos, pos - numpos).c_str());
		}
	      }
	    }
	  }
	}
      }

      stringstream ss;
      ss << llround (cputime [vedi->jobExeID] * freq * REALSI2000PERMHZ);

      xmlNodePtr cpu = xmlNewChild (jed, NULL, BAD_CAST "cpuConsumption", NULL);
      xmlNewChild (cpu, NULL, BAD_CAST "count", BAD_CAST ss.str().c_str());
      xmlNewChild (cpu, NULL, BAD_CAST "unit", BAD_CAST "specint2000seconds");
    }

    if (poolfile.find (vedi->jobExeID) != poolfile.end()) {

      xmlInitParser();

      xmlDocPtr pooldoc = xmlParseMemory (poolfile [vedi->jobExeID].c_str(),
					  poolfile [vedi->jobExeID].length());
      if (!pooldoc) return NULL;

      xmlNodePtr poolroot = xmlDocGetRootElement (pooldoc);

      xmlNodePtr fmd = xmlNewChild (jed, NULL, BAD_CAST "fileMetaData", NULL);
      xmlAddChild (fmd, xmlCopyNode (poolroot, true));

      xmlFreeDoc (pooldoc);
    }

    bool rlsfailed = false;
    if (xrsl.find (vedi->jobExeID) != xrsl.end()) {

      xmlNodePtr opf = xmlNewChild (jed, NULL, BAD_CAST "jobOutputs", NULL);

      globus_rsl_t * axrsl =
	globus_rsl_parse ((char *) xrsl [vedi->jobExeID].c_str());

      globus_rsl_t * outputfilesrelation = NULL;
      FindRelation ("outputfiles", &outputfilesrelation, axrsl);

      if (outputfilesrelation) {

	globus_rsl_value_t * sequence =
	  globus_rsl_relation_get_value_sequence (outputfilesrelation);
	if (!globus_rsl_value_is_sequence (sequence))
	  cerr << "Error: XRSL syntax error in attribute \"outputfiles\"" << endl;

	globus_list_t * list1 =
	  globus_rsl_value_sequence_get_value_list (sequence);

	while (!globus_list_empty (list1)) {

	  globus_rsl_value_t * value1 =
	    (globus_rsl_value_t *) globus_list_first (list1);
	  if (!globus_rsl_value_is_sequence (value1))
	    cerr << "Error: XRSL syntax error in attribute \"outputfiles\"" << endl;

	  globus_list_t * list2 =
	    globus_rsl_value_sequence_get_value_list (value1);

	  int paircnt = 0;
	  string lfn;
	  string pfn;

	  while (!globus_list_empty (list2)) {

	    globus_rsl_value_t * value2 =
	      (globus_rsl_value_t *) globus_list_first (list2);

	    switch (paircnt) {

	    case 0:
	      if (globus_rsl_value_is_literal (value2))
		lfn = globus_rsl_value_literal_get_string (value2);
	      break;

	    case 1:
	      if (globus_rsl_value_is_literal (value2)) {
                pfn = globus_rsl_value_literal_get_string (value2);
                int loc1 = pfn.find ("://");
                if (loc1 != string::npos) {
                  int loc2 = pfn.find ('@', loc1);
                  if (loc2 != string::npos)
                    pfn.erase (loc1 + 3, loc2 - loc1 - 2);
                }
	      }
	      break;
	    }

	    paircnt++;
	    list2 = globus_list_rest (list2);
	  }
	  if (paircnt != 2)
	    cerr << "Error: XRSL syntax error in attribute \"outputfiles\"" << endl;

	  bool vers = true;
	  int ptpos = pfn.rfind ('.');
	  for (int i = pfn.length() - 1; i > ptpos; i--) {
	    if (!isdigit (pfn[i])) {
	      vers = false;
	      break;
	    }
	  }
	  if (vers) pfn = pfn.substr (0, ptpos);

	  if (!pfn.empty ()) {
	    xmlNodePtr fin = xmlNewChild (opf, NULL, BAD_CAST "fileInfo", NULL);
	    xmlNewChild (fin, NULL, BAD_CAST "LFN", BAD_CAST lfn.c_str());
	    xmlNodePtr tul = xmlNewChild (fin, NULL, BAD_CAST "turlList", NULL);
	    xmlNewChild (tul, NULL, BAD_CAST "turl", BAD_CAST pfn.c_str());
	    xmlNewChild (fin, NULL, BAD_CAST "MD5sums",
			 BAD_CAST metadata [vedi->jobExeID][lfn].md5sum.c_str());
	    xmlNewChild (fin, NULL, BAD_CAST "size",
			 BAD_CAST metadata [vedi->jobExeID][lfn].size.c_str());
	    xmlNewChild (fin, NULL, BAD_CAST "logCol",
			 BAD_CAST metadata [vedi->jobExeID][lfn].lcn.c_str());
	    xmlNodePtr dsn = xmlNewChild (fin, NULL, BAD_CAST "dataset", NULL);
	    xmlNewChild (dsn, NULL, BAD_CAST "name",
			 BAD_CAST metadata [vedi->jobExeID][lfn].dataset.c_str());

	    if (pfn.substr (0, 6) == "rls://") {
	      int pos = pfn.find ('/', 6);

	      int errc;
	      char errmsg [512];

	      globus_result_t res;
	      globus_rls_handle_t * rlshandle;
	      res = globus_rls_client_connect ((char *) pfn.substr (0, pos + 1).c_str(),
					       &rlshandle);
	      if (res) {
		globus_rls_client_error_info (res, &errc, errmsg, 512, false);
		cout << errmsg << endl;
		rlsfailed = true;
	      }

	      else {
		for (int i = 0; i < 5; i++) {

		  globus_rls_attribute_t attr;
		  attr.val.s = NULL;

		  switch (i) {
		  case 0:
		    attr.name = "dq_guid";
		    attr.val.s = strdup (metadata [vedi->jobExeID][lfn].guid.c_str());
		    break;
		  case 1:
		    attr.name = "dq_date";
		    attr.val.s = strdup (metadata [vedi->jobExeID][lfn].date.c_str());
		    break;
		  case 2:
		    attr.name = "dq_lcn";
		    attr.val.s = strdup (metadata [vedi->jobExeID][lfn].lcn.c_str());
		    break;
		  case 3:
		    attr.name = "dq_fsize";
		    attr.val.s = strdup (metadata [vedi->jobExeID][lfn].size.c_str());
		    break;
		  case 4:
		    attr.name = "dq_md5sum";
		    attr.val.s = strdup (metadata [vedi->jobExeID][lfn].md5sum.c_str());
		    break;
		  }

		  attr.objtype = globus_rls_obj_lrc_lfn;
		  attr.type = globus_rls_attr_type_str;

		  if (attr.val.s && *attr.val.s) {
		    res = globus_rls_client_lrc_attr_add (rlshandle,
							  (char *) pfn.substr (pos + 1).c_str(),
							  &attr);
		    if (res) {
		      globus_rls_client_error_info (res, &errc, errmsg, 512, false);
		      cout << errmsg << endl;
		    }
		  }

		  if (attr.val.s) free (attr.val.s);
		}

		res = globus_rls_client_close (rlshandle);
		if (res) {
		  globus_rls_client_error_info (res, &errc, errmsg, 512, false);
		  cout << errmsg << endl;
		}
	      }
	    }
	  }

	  list1 = globus_list_rest (list1);
	}
      }
    }

    string jobstatus = "unknown";
    if (status [vedi->jobExeID].find ("ACCEPTED") != string::npos)
      jobstatus = "pending";
    else if (status [vedi->jobExeID].find ("PREPARING") != string::npos ||
	     status [vedi->jobExeID].find ("INLRMS") != string::npos ||
	     status [vedi->jobExeID].find ("FINISHING") != string::npos)
      jobstatus = "running";
    else if (status [vedi->jobExeID].find ("FINISHED") != string::npos) {
      if (xrsl.find (vedi->jobExeID) != xrsl.end())
	jobstatus = "finished";
      if (!error [vedi->jobExeID].empty()) jobstatus = "failed";
      // if the RLS server is down the DQ attributes have not been registered
      // so we have to trick the supervisor to ask again by pretending the
      // job is still running
      if (rlsfailed) jobstatus = "running";
    }
    else if (status [vedi->jobExeID].find ("DELETED") != string::npos)
      jobstatus = "failed";
    xmlNewChild (jed, NULL, BAD_CAST "jobStatus", BAD_CAST jobstatus.c_str());

    xmlNewChild (jed, NULL, BAD_CAST "jobNativeStatus",
		 BAD_CAST status [vedi->jobExeID].c_str());

    if (!error [vedi->jobExeID].empty())
      xmlNewChild (jed, NULL, BAD_CAST "validationDiagnostic",
		   BAD_CAST error [vedi->jobExeID].c_str());

    if (!host [vedi->jobExeID].empty())
      xmlNewChild (jed, NULL, BAD_CAST "host",
		   BAD_CAST host [vedi->jobExeID].c_str());

    int pos1 = error [vedi->jobExeID].find ("JOB FAILURE: Job exit code is ");
    if (pos1 != string::npos) {
      pos1 += 30;
      int pos2 = error [vedi->jobExeID].find (" != 0", pos1);
      xmlNewChild (jed, NULL, BAD_CAST "jobExitCode",
		   BAD_CAST error [vedi->jobExeID].substr (pos1, pos2 - pos1).c_str());
    }
    else if (status [vedi->jobExeID].find ("FINISHED") != string::npos)
      xmlNewChild (jed, NULL, BAD_CAST "jobExitCode", BAD_CAST "0");
  }

  globus_module_deactivate (GLOBUS_RSL_MODULE);
  globus_module_deactivate (GLOBUS_RLS_CLIENT_MODULE);
  globus_module_deactivate (GLOBUS_IO_MODULE);
  globus_module_deactivate (GLOBUS_COMMON_MODULE);

  xmlChar * x;
  int len;

  xmlDocDumpFormatMemoryEnc (doc, &x, &len, "utf-8", 0);

  xmlFreeDoc (doc);

  if (debug)
    cout << "xml passed by executor: " << x << endl;

  return (char *) x;
}

char * fixJobs (char * xml) {

  if (debug)
    cout << "xml passed by supervisor: " << xml << endl;

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

  vector <jobExeData> exdata;

  while (xmlnode1) {
    if (strcmp ((const char *) xmlnode1->name, "killJob") == 0) {
      xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
      while (xmlnode2) {
	if (strcmp ((const char *) xmlnode2->name, "jobExeData") == 0) {
	  xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	  jobExeData jobDef;
	  while (xmlnode3) {
	    if (strcmp ((const char *) xmlnode3->name, "jobDefID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobDefID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobName") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobName =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobExeID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobExeID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "infoExecutor") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.infoExecutor =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "action") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.action =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    xmlnode3 = xmlnode3->next;
	  }
	  exdata.push_back (jobDef);
	}
	xmlnode2 = xmlnode2->next;
      }
    }
    xmlnode1 = xmlnode1->next;
  }

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  vector <string> cleanids;
  vector <string> restageids;
  vector <string> alertids;
  for (vector <jobExeData>::const_iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++) {
    if (vedi->action == "clean")
      cleanids.push_back (vedi->jobExeID);
    if (vedi->action == "restage")
      restageids.push_back (vedi->jobExeID);
    if (vedi->action == "alert")
      alertids.push_back (vedi->jobExeID);
  }

  stringstream s, dummy;
  streambuf * sbo = cout.rdbuf(s.rdbuf());
  streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
  vector <string> e;
  string f;
  if (!cleanids.empty())
    ngcleanxx (cleanids, e, e, e, e, false, false, 20, 0, true);
  if (!restageids.empty())
    ngresubxx (restageids, e, e, e, e, false, e, e, e, f, false, false, false,
	       20, 0, true);
  if (!alertids.empty())
    ngrenewxx (alertids, e, e, e, e, false, 20, 0, true);
  cerr.rdbuf(sbe);
  cout.rdbuf(sbo);

  return NULL;
}

char * killJobs (char * xml) {

  if (debug)
    cout << "xml passed by supervisor: " << xml << endl;

  xmlInitParser();

  xmlDocPtr xmldoc = xmlParseMemory (xml, strlen (xml));
  if (!xmldoc) return NULL;

  xmlNodePtr xmlroot = xmlDocGetRootElement (xmldoc);

  xmlNodePtr xmlnode1 = xmlroot->xmlChildrenNode;

  vector <jobExeData> exdata;

  while (xmlnode1) {
    if (strcmp ((const char *) xmlnode1->name, "killJob") == 0) {
      xmlNodePtr xmlnode2 = xmlnode1->xmlChildrenNode;
      while (xmlnode2) {
	if (strcmp ((const char *) xmlnode2->name, "jobExeData") == 0) {
	  xmlNodePtr xmlnode3 = xmlnode2->xmlChildrenNode;
	  jobExeData jobDef;
	  while (xmlnode3) {
	    if (strcmp ((const char *) xmlnode3->name, "jobDefID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobDefID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobName") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobName =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "jobExeID") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.jobExeID =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    if (strcmp ((const char *) xmlnode3->name, "infoExecutor") == 0)
	      if (xmlnode3->xmlChildrenNode)
		jobDef.infoExecutor =
		  (const char *) xmlnode3->xmlChildrenNode->content;
	    xmlnode3 = xmlnode3->next;
	  }
	  exdata.push_back (jobDef);
	}
	xmlnode2 = xmlnode2->next;
      }
    }
    xmlnode1 = xmlnode1->next;
  }

  xmlFreeDoc (xmldoc);
  xmlCleanupParser();

  vector <string> jobids;
  for (vector <jobExeData>::const_iterator vedi = exdata.begin();
       vedi != exdata.end(); vedi++)
    jobids.push_back (vedi->jobExeID);

  stringstream s, dummy;
  streambuf * sbo = cout.rdbuf(s.rdbuf());
  streambuf * sbe = cerr.rdbuf(dummy.rdbuf());
  vector <string> e;
  ngkillxx (jobids, e, e, e, e, false, false, 20, 0, true);
  cerr.rdbuf(sbe);
  cout.rdbuf(sbo);

  return NULL;
}
